# Matrix implementation wrappers.
#
# Placed in the public domain.
#
# April 05, 2000
# Neil Schemenauer <nascheme@enme.ucalgary.ca>

import sys

MatrixError = 'MatrixError'

class AbstractMatrix:
    def __getitem__(self, index):
        e = apply(self.get_element, index)
        return e.get()

    def __setitem__(self, index, v):
        apply(self.set_element, index+(v,))

#
# A sparse matrix object implemented in C.  This C library used is Kenneth S.
# Kundert's "Sparse".
#
class SparseMatrix(AbstractMatrix):
	def __init__(self, size):
		import sparse
		self.matrix = sparse.sparse(size)

	# write the matrix to file
	def write(self, file=sys.stdout):
		self.matrix.write(file, 0, 1, 1)

	# set all elements to 0
	def clear(self):
		self.matrix.clear()

	# get a element object to element (i, j), supports .get and .set
	def get_element(self, i, j):
		return self.matrix.get_element(i, j)

        def set_element(self, i, j, v):
                self.matrix.set_element(i, j, v)

	# solve matrix and return solution vector (as a list)
	def solve(self, RHS):
		try:
			self.matrix.factor()
			return self.matrix.solve(RHS)
		except sparse.error, value:
			raise MatrixError, value


#
# Use PyNum (Python Numerical Extensions).  This is _not_ a sparse
# matrix.  For many equations this will waste a lot of space.  It is
# OK for small systems though.  Also, I think the linear equation
# solver is quite robust (it is from LAPACK).
#
class NumericMatrix(AbstractMatrix):
	def __init__(self, size):
		self.size = size
		self.clear()

	def clear(self):
		self.matrix = Numeric.zeros((self.size, self.size), 
                                            Numeric.Float64)
	def write(self, file=sys.stdout):
		file.write(`self.matrix`)

	def get_element(self, i, j):
		return NumericElement(self, i, j)

        def set_element(self, i, j, v):
                self.matrix[i][j] = v

	def solve(self, RHS):
		RHS = Numeric.array(RHS)
		return LinearAlgebra.solve_linear_equations(self.matrix, RHS)

#
# Since NumPy matrix objects can't return an element reference object
# we have to make our own.
#
class NumericElement:
	def __init__(self, matrix, i, j):
		self.matrix = matrix
		self.i = i
		self.j = j

	def set(self, val):
		self.matrix.matrix[self.i, self.j] = val

	def get(self):
		return self.matrix.matrix[self.i, self.j]


class SimpleMatrix(AbstractMatrix):
	def __init__(self, size):
		self.a = [None] * size
		for i in range(size):
			self.a[i] = [0.0]*size

	def clear(self):
		for i in range(len(self.a)):
			for j in range(len(self.a)):
				self.a[i][j] = 0.0
		
	def get_element(self, i, j):
		return SimpleElement(self.a, i, j)

	def set_element(self, i, j, v):
                self.a[i][j] = v

	def solve(self, RHS):
		import LUP
		pi = LUP.lup_decomposition(self.a)
		x = LUP.lup_solve(self.a, self.a, pi, RHS)
		return x

class SimpleElement:
	def __init__(self, a, i, j):
		self.a = a
		self.i = i
		self.j = j

	def set(self, v):
		self.a[self.i][self.j] = v

	def get(self):
		return self.a[self.i][self.j]


#
# Use sparsemodule if we can, otherwise fall back to NumPy
#
try:
	import sparse
	Matrix = SparseMatrix
except ImportError:
	print 'No sparse, using NumPy'
	try:
		import LinearAlgebra, Numeric
		Matrix = NumericMatrix
	except ImportError:
		print 'No NumPy, using very slow SimpleMatrix'
		Matrix = SimpleMatrix
